#include "FATParser.h"
#include "../common.h"
#include "../filesystem/File.h"
#include <string>

FATParser::FATParser(FILE* rawDisk, unsigned long long partitionOffset) : FilesystemParser(rawDisk, partitionOffset) {
  unsigned char buffer[512];
  fseeko(rawDisk, partitionOffset, SEEK_SET);
  fread(buffer, 1, 512, rawDisk);
  this->device = rawDisk;

  sectorsPerCluster    = buffer[13];
  reservedSectors      = get_word(&buffer[14]);
  fileAllocationTables = buffer[16] * get_dword(&buffer[36]);
  firstSector          = reservedSectors + fileAllocationTables;
  partitionStart       = partitionOffset;
}

unsigned long long FATParser::FATE(unsigned long long cluster) {
  unsigned long long fatOffset = cluster * 4;       // 4 byte FAT entries for FAT32
  unsigned long long fatSector = fatOffset >> 9;    // div by 512, bytes per sector
  fatOffset = fatOffset - (fatSector << 9);         // times by 512, to get offset in sector

  unsigned long long pos = partitionStart + ((reservedSectors + fatSector) * 512);

  unsigned char buffer[512];
  fseeko(device, pos, SEEK_SET);
  fread(buffer, 1, 512, device);

  // return the next cluster in the file
  unsigned long long result = get_dword(buffer);
  return result & 0x0FFFFFFF;
}

string FATParser::getFilename(unsigned char* buff) {
  string filename = "";

  // add the first 8 bytes
  for(int i = 0; i < 8; i++) {
    if(buff[i] == 0) {
      break;
    }

    filename += (char)buff[i];
  }

  if(filename[0] == '.') {
    return filename;
  }

  // add the file extension
  string extension = "";
  for(int i = 0; i < 3; i++) {
    if(buff[i + 8] == 0) {
      break;
    }

    extension += (char)buff[i + 8];
  }

  if(extension != "   ") {
    filename += ".";
    filename += extension;
  }

  return filename;
}

void FATParser::enumerateNode(Directory* node, unsigned long long cluster, unsigned long long pos) {
  unsigned long long lastCluster = cluster;
  printf("Enumerating node.\n");

  while(true) {
    // enumerate for all entries in a directory sector
    for(int j = 0; j < sectorsPerCluster; j++) {
      unsigned char buffer[512];
      fseeko(device, pos, SEEK_SET);
      fread(buffer, 1, 512, device);

      if(buffer[0] == 0x00) {
        break;
      }

      for(int i = 0; i < 16; i++) {
        string filename = "";

        if((buffer[(i*32)+11] & 0x08) == 0x08) {
          continue;
        }
        if(buffer[i*32] == 0x00) { // file doesn't exist
          break;
        }

        if(buffer[i*32] == 0xE5) { // deleted file
          filename += "?";
        }
        else {
          filename = getFilename(&buffer[i*32]);
        }

        // it's not the current directory or the parent directory
        if(filename[0] != '.') {
          if(buffer[i*32] == 0xE5) {
            // handle deleted files
          }
          else {
            // create new node
            if((buffer[(i*32)+11] & 0x10) == 0x10) {
              unsigned long long subCluster = ((get_word(&buffer[(i*32)+20]) << 16) + get_word(&buffer[(i*32)+26]));
              unsigned long long subPos = partitionStart + ((firstSector + ((subCluster-2)*sectorsPerCluster))*512);
              Directory* dir = new Directory(filename.c_str(), NULL);
              enumerateNode(dir, subCluster, subPos);
            }
            else {
              File* f = new File(filename.c_str(), NULL);
              node->addNode(f);
            }
          }
        }
      }
      pos += 512;
    }

    unsigned long long nextCluster = FATE(lastCluster);
    printf("Last cluster: %llu, Next Cluster: %llu\n", lastCluster, nextCluster);
    if(nextCluster >= 0x0FFFFFF8) {
      break;
    }
    lastCluster = nextCluster;
    pos = partitionStart + ((firstSector + ((nextCluster-2) * sectorsPerCluster)) * 512);
  }
}

Directory* FATParser::getRoot() {
  Directory* root = new Directory(".", NULL);
  unsigned long long rootDir = partitionStart + ((reservedSectors + fileAllocationTables)*512);
  enumerateNode(root, 2, rootDir);
  return root;
}

