#include "Attribute.h"
#include <stdio.h>
#include <stdlib.h>

#include "DataRun.h"
#include "FilenameAttribute.h"
#include "IndexRootAttribute.h"
#include "IndexAllocationAttribute.h"
#include "DataAttribute.h"
#include "StandardInformationAttribute.h"

NTFSAttribute* getNTFSAttribute(DeviceInformation* device, byte* offset, int mftRecordOffset) {
  // identify the attribute type, and create the appropriate subclass of attribute
  int type = get_dword(&offset[0x00]);
  if(type == 0x10)
    return new NTFSStandardInformationAttribute(device, offset, mftRecordOffset);
  else if(type == 0x30)
    return new NTFSFilenameAttribute(device, offset, mftRecordOffset);
  else if(type == 0x80)
    return new NTFSDataAttribute(device, offset, mftRecordOffset);
  else if(type == 0x90)
    return new NTFSIndexRootAttribute(device, offset, mftRecordOffset);
  else if(type == 0xA0)
    return new NTFSIndexAllocationAttribute(device, offset, mftRecordOffset);
  else
    return new NTFSAttribute(device, offset, mftRecordOffset);
}

NTFSAttribute::NTFSAttribute(DeviceInformation* device, unsigned char* offset, int mftRecordOffset) {
    int intType = get_dword(&offset[0x00]);
    valid = true;

    if(intType == 0xFFFF)  { // End marker
        valid = false;
        return;
    }
    else {
        setType(intType);
    }

    this->device = device;


    length          = get_dword(&offset[0x04]);
    nonResident     = (int)offset[0x08];
    lengthOfName    = (int)offset[0x09];
    offsetToName    = get_word(&offset[0x0a]);
    flags           = get_word(&offset[0x0c]);
    attributeID     = get_word(&offset[0x0e]);
    sizeOfContent   = get_dword(&offset[0x10]);
    offsetToContent = get_word(&offset[0x15]);
    
    if(lengthOfName != 0 && lengthOfName + offsetToName < mftRecordOffset + 1024) {
        name = (char*)malloc(lengthOfName + 1);
        for(int i = 0; i < lengthOfName; i++) {
          name[i] = offset[offsetToName + (i*2)];
        }
        name[lengthOfName] = 0;
    }
    else {
        name = NULL;
    }

    // create any data runs, if they exist
    if(nonResident == 1) {
      int dataRunStart = 0x40;
      
      if(offsetToName != 0) {
        dataRunStart += (lengthOfName * 2);  // times two for unicode
      }

      //printf("Data run starts at (type: %X): %d\n", dataRunStart, intType);

      NTFSDataRun* run = new NTFSDataRun(offset + dataRunStart, 0);

      while(run->getLength() != 0) {
        dataRuns.push_back(run);

        dataRunStart += (run->getRunLength());
        run = new NTFSDataRun(offset + dataRunStart, run->getLCNOffset());
      }

      // delete the last (0-length) run
      delete run;
    }
}

NTFSAttribute::~NTFSAttribute() {
    if(valid && name != NULL)
        free(name);

    for(unsigned int i = 0; i < dataRuns.size(); i++) {
      delete dataRuns[i];
    }
}

vector<NTFSDataRun*> NTFSAttribute::getDataRuns() {
  return dataRuns;
}

bool NTFSAttribute::isValid() {
    return valid;
}

int NTFSAttribute::getLength() {
    return length;
}

int NTFSAttribute::getHeaderLength() {
  if(getResident() == true) {
    if(hasName()) {
      return 0x18 + (2*lengthOfName);
    }
    else {
      return 0x18;
    }
  }
  else {
    if(hasName()) {
      return 0x40 + (2*lengthOfName);
    }
    else {
      return 0x40;
    }
  }
}

void NTFSAttribute::setType(int intType) {
    if(intType % 16 != 0 || intType / 16 > 16 || intType <= 0) {
        type = INVALID;
    }
    else {
        type = (NTFSAttributeType)(intType / 16);
    }
}

NTFSAttributeType NTFSAttribute::getType() {
    return type;
}

bool NTFSAttribute::hasName() {
    return lengthOfName != 0;
}

int NTFSAttribute::getNameLength() {
  return lengthOfName;
}

const char* NTFSAttribute::getName() {
  return name;
}

bool NTFSAttribute::getResident() {
  return nonResident == 0;
}

NTFSData NTFSAttribute::getData() {
  unsigned long long length = 0;

  for(unsigned int i = 0; i < dataRuns.size(); i++) {
    length += dataRuns[i]->getLength();
  }

  NTFSData data = NTFSData(length*device->clusterSize);

  for(unsigned int i = 0; i < dataRuns.size(); i++) {
    unsigned char* buf = (unsigned char*)malloc(dataRuns[i]->getLength() * device->clusterSize);

    fseeko(device->rawDevice, (dataRuns[i]->getLCNOffset()*device->clusterSize) + device->partitionOffset, SEEK_SET);
    fread(buf, 1, (dataRuns[i]->getLength()*device->clusterSize), device->rawDevice);
    data.append(buf, dataRuns[i]->getLength() * device->clusterSize);

    free(buf);
  }

  return data;
}

bool NTFSAttribute::containsDataRuns() {
  return dataRuns.size() != 0;
}

void NTFSAttribute::print() {
    switch(type * 16) {
        case 0x10:  // standard attributes
            printf("Standard attributes");
            break;
        case 0x20:  // attribute list
            printf("Attribute list");
            break;
        case 0x30:  // filename
            printf("Filename");
            break;
        case 0x40:  // volume info or object_id
            printf("Volume info or object id");
            break;
        case 0x50:
            printf("Security descriptor");
            break;
        case 0x60:
            printf("Volume name");
            break;
        case 0x70:
            printf("Volume information");
            break;
        case 0x80:
            printf("Data");
            break;
        case 0x90:
            printf("Index Root");
            break;
        case 0xA0:
            printf("Index allocation");
            break;
        case 0xB0:
            printf("Bitmap (length: %d)", length);
            break;
        case 0xC0:
            printf("Symbolic link / reparse point");
            break;
        case 0xD0:
            printf("EA Information");
            break;
        case 0xE0:
            printf("EA");
            break;
        case 0xF0:
            printf("Property set");
            break;
        case 0x100:
            printf("Logged utility stream");
            break;
        case 0xFFFF:
            printf("End attribute");
            break;
        default:
            printf("Unknown attribute type");
    }

    if(hasName()) {
      printf(" (%s)\n", getName());
    }
    else {
      printf("\n");
    }
}

