#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>

#include "../common.h"
#include "Attribute.h"

#include "NTFSParser.h"


NTFSParser::NTFSParser(FILE* rawDisk, unsigned long long partitionOffset)
  : FilesystemParser(rawDisk, partitionOffset) {
  unsigned char sector[512];

  bytesPerSector           = 0;
  sectorsPerClustor        = 0;
  mediaDescriptor          = 0;
  sectorsPerTrack          = 0;
  numberOfHeads            = 0;
  hiddenSectors            = 0;
  totalSectors             = 0;
  logicalClustorForMFT     = 0;
  logicalClustorForMFTMirr = 0;

  device = new DeviceInformation;

  device->partitionOffset = partitionOffset;
  device->rawDevice = rawDisk;

  fseek(rawDisk, partitionOffset, SEEK_SET);
  fread(sector, 1, 512, rawDisk);

  bytesPerSector    = get_word(&sector[0x0B]);
  sectorsPerClustor = sector[0x0D];
  mediaDescriptor   = sector[0x15];
  sectorsPerTrack   = get_word(&sector[0x18]);
  numberOfHeads     = get_word(&sector[0x1A]);
  hiddenSectors     = get_dword(&sector[0x1C]);
  totalSectors      = get_longlong(&sector[0x28]);

  device->clusterSize = bytesPerSector * sectorsPerClustor;

  logicalClustorForMFT     = get_longlong(&sector[0x30]);
  logicalClustorForMFTMirr = get_longlong(&sector[0x38]);
   
  unsigned long long mft_offset = device->clusterSize*logicalClustorForMFT;

  // read the MFT entry for the MFT itself
  unsigned char rawMft[1024];
  fseeko64(device->rawDevice, mft_offset, SEEK_SET);
  fread(rawMft, 1, 1024, device->rawDevice);
  MFTEntry* mftmftentry = new MFTEntry(device, rawMft);
  mftmftentry->print();

  mftData = mftmftentry->getData();
  mft = mftData.getData();
  mftLength = mftData.getLength();
  
  delete mftmftentry;
}

NTFSParser::~NTFSParser() {
	if(mft != NULL) {
		free(mft);
	}
}

Node* NTFSParser::getNode(unsigned long long mftEntryNumber, Node* parent) {
	MFTEntry* entry = getMFTEntry(mftEntryNumber);
	Node* returningNode = NULL;
	if(entry == NULL) {
		return NULL;
	}
	
	const char* filename = entry->getFilename();
	
	if(filename == NULL || filename[0] == '$') {
		delete entry;
		return NULL;
	}
	
	if(!entry->isDirectory()) {
		returningNode = new File(filename, parent);
	}
	else {
		Directory* dir = new Directory(filename, parent);
		
		vector<IndexEntry> entries = entry->getIndexEntries();
		for(unsigned int i = 0; i < entries.size(); i++) {
			if(entries[i].getFileReference() == mftEntryNumber) {
				continue;
			}
			
			Node* node = getNode(entries[i].getFileReference(), dir);

			if(node != NULL) {
				dir->addNode(node);
			}
		}
		
		returningNode = dir;
	}
	
	delete entry;
	return returningNode;
}

Directory* NTFSParser::getRoot() {
	return (Directory*)getNode(5, NULL);
}

MFTEntry* NTFSParser::getMFTEntry(unsigned long long mftEntryNumber) {
	if(mftEntryNumber >= (mftLength/1024)) {
		fprintf(stderr, "ERROR: MFT Entry %llu is not in the MFT!\n", mftEntryNumber);
		return NULL;
	}
	else {
		return new MFTEntry(device, &mft[mftEntryNumber * 1024]); 
	}
}

